/*  GNU XX, graphics display optimization tools
    Copyright (C) 1992 Free Software Foundation, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef RTREEH
#define RTREEH
/*  t. lord	Wed Feb 12 22:42:27 1992	*/

#include "ir.h"
#include "pile.h"

typedef struct xx_sRtree * xx_Rtree;
typedef struct xx_sRtreeNode * xx_RtreeNode;
typedef struct xx_sRtreeLeaf * xx_RtreeLeaf;
struct xx_sRtreeLeaf
{
  xx_RtreeNode parent;		/* must match xx_RtreeItem */
  int pos;			/* must match xx_RtreeItem */
  void *data;
};

#ifdef __STDC__
typedef int (*xx_LoopFn) (xx_RtreeLeaf, xx_IntRectangle, void *loopstate);
#else
typedef int (*xx_LoopFn) ();
#endif

typedef ut_Pile xx_RtreeLeafPile;

#define xx_rtreeleafpile()  ut_pile( sizeof (xx_RtreeLeaf), 10 )
#define xx_rtreeleafpile_read(P)  (xx_RtreeLeaf*)(P)->data
#define xx_rtreeleafpile_add(P,L) ((void)\
				   (*(xx_RtreeLeaf*)ut_pile_reserve((P),1)\
				    = (L)))
#define xx_rtreeleafpile_clear ut_pile_clear
#define xx_free_rtreeleafpile ut_free_pile
#define xx_rtreeleafpile_size(P) ((P)->used)

struct xx_sRtree 
{
  int depth;
  struct xx_sIntRectangle domain;
  xx_RtreeNode root;
};

#define RTREE_FANOUT 3

struct xx_sRtreeNode
{
  xx_RtreeNode parent;		/* must match xx_RtreeItem */
  int pos;			/* must match xx_RtreeItem */
  struct xx_sIntRectangle rectangles[RTREE_FANOUT];
  void * children[RTREE_FANOUT];
};


typedef struct xx_sRtreeItem * xx_RtreeItem;
struct xx_sRtreeItem
{
  xx_RtreeNode parent;
  int pos;
};

#define xx_rtreenode_subtree(NODE,X)	\
	((X) < RTREE_FANOUT ? (xx_RtreeNode)(NODE)->children[X] : 0)
#define xx_rtreenode_leaf(NODE,X)	\
	((X) < RTREE_FANOUT ? (xx_RtreeLeaf)(NODE)->children[X] : 0)
#define xx_rtreenode_item(NODE,X)	\
	((X) < RTREE_FANOUT ? (xx_RtreeItem)(NODE)->children[X] : 0)



#ifdef __STDC__
extern xx_RtreeLeaf xx_rtreeleaf (void * data);
extern void xx_free_rtreeleaf (xx_RtreeLeaf);
extern xx_Rtree xx_rtree ();
extern void xx_free_rtree (xx_Rtree);
extern int xx_rtree_empty (xx_Rtree);
extern void xx_rtree_clear (xx_Rtree);
extern void xx_rtree_delete (xx_Rtree, xx_RtreeLeaf);
extern void xx_rtree_query (xx_Rtree, xx_IntRectangle,
			      xx_LoopFn, void *);
extern void xx_rtree_add (xx_Rtree, xx_IntRectangle, xx_RtreeLeaf);
extern xx_Rtree xx_rtree_copy (xx_Rtree, xx_LoopFn leaf_copier);
extern void xx_rtree_translate (xx_Rtree, int dx, int dy);
#else
extern xx_RtreeLeaf xx_rtreeleaf ();
extern void xx_free_rtreeleaf ();
extern xx_Rtree xx_rtree ();
extern xx_Rtree xx_free_rtree ();
extern int xx_rtree_empty ();
extern void xx_rtree_clear ();
extern void xx_rtree_delete ();
extern void xx_rtree_query ();
extern void xx_rtree_add ();
extern xx_Rtree xx_rtree_copy ();
extern void xx_rtree_translate ();
#endif


#ifdef PRETTY_PRINT_DECLS
#ifdef __STDC__
extern void xx_pp_rtree (ut_PPstate, xx_Rtree, ut_PPfn leaffn);
#else
extern void xx_pp_rtree ();
#endif
#define xx_pp_rtree_name(S,F) ut_pp_name((F),"xx_Rtree",(S))
#endif
#endif

#ifdef PRETTY_PRINT
#define xx_pp_rtreenode_name(S,F) ut_pp_name("xx_RtreeNode",(S), (F))

void
xx_pp_rtreenode (state, node, depth, leaffn)
     ut_PPstate state;
     xx_RtreeNode node;
     int depth;
     ut_PPfn leaffn;
{
  char *indent = ut_indent (state->indentation);

  if (node)
    {
      int x;
      fprintf (state->stream, "%s(xx_RtreeNode %x  -- pos %d, parent ",
	       indent, node, node->pos);
      xx_pp_rtreenode_name (node->parent, state->stream);
      fputc ('\n', state->stream);
      state->indentation += 2;
      for (x = 0; x < RTREE_FANOUT && node->children[x]; ++x)
	{
	  fprintf (state->stream, "%s child %d @ ", indent, x);
	  xx_pp_intrectangle (node->rectangles + x, state->stream);
	  fputc('\n', state->stream);
	  if (depth > 1)
	    xx_pp_rtreenode (state, node->children[x], depth - 1, leaffn);
	  else
	    leaffn (state, xx_rtreenode_leaf(node, x)->data);
	}
      state->indentation -= 2;
    }
  else
    {
      fprintf (state->stream, "%s(xx_RtreeNode 0)\n", indent);
    }
}

void
xx_pp_rtree (state, rtree, leaffn)
     ut_PPstate state;
     xx_Rtree rtree;
     ut_PPfn leaffn;
{
  char * indent = ut_indent (state->indentation);

  if (rtree)
    {
      fprintf (state->stream, "%s(xx_Rtree %x  -- depth %d,\n%s domain ",
	       indent, rtree, rtree->depth, indent);
      xx_pp_intrectangle (&rtree->domain, state->stream);
      fputc ('\n', state->stream);
      state->indentation += 2;
      xx_pp_rtreenode (state, rtree->root, rtree->depth, leaffn);
      fprintf (state->stream, "%sxx_Rtree %x)\n", indent, rtree);
    }
  else
    fprintf (state->stream, "%s(xx_Rtree 0)\n", indent);
}



#endif
